function microfract3D

%%%%%%%%%%%%%%%%%%
%3d crack problem...given three points defining a macro crack plane on an image of cohesive energies, find the
%most likely crack on the microstructural scale
%    CLASS:: An 1xN vector which specifies the initial labels of each
%      of the N nodes in the graph
%    UNARY:: A CxN matrix specifying the potentials (data term) for
%      each of the C possible classes at each of the N nodes. This is the
%      penalty term
%    PAIRWISE:: An NxN sparse matrix specifying the graph structure and
%      cost for each link between nodes in the graph. 
%    LABELCOST:: A CxC matrix specifying the fixed label cost for the
%      labels of each adjacent node in the graph. 

clear
close all

%%%%%%%%%%%%%%%%%%RUN EXAMPLES FROM THE COLOR VOL FILE%%%%%%%%%%%%%%%%%%
n = 64;
maxlevels = 2;
fid = fopen('particles.vol', 'r', 'l');
mydata = fread(fid, inf, 'uchar',0,'l');
s = mydata(4097:end);
RGB = uint8(reshape(s,3,n^3)');
RGB2 = rgb2ind(reshape(RGB,n^3,1,3),maxlevels);
bondfull = double(reshape(RGB2,n,n,n));%particles are hard to cut 1, matrix soft 0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%USER INPUTS IN MATRIX, COMMENT PREVIOUS BLOCK (lines 20-27) AND UNCOMMENT THIS BLOCK (lines 32-34) IF YOU ARE USING YOUR OWN 3D MICROSTRUCTURE%%%%%%%%%%%%%%%%
%%%%%%%%%%NOTE: The microstructure is in a 3D matrix (of size n^3) is saved as "bondfull" (see 3dimage.mat). enter the correct maxlevels
% load 3dimage
% n = length(bondfull);
% maxlevels = 2;

%%%%%%%%USER can specify the size reduction for the microstructure in variable 'mag' (from n^3 to (n*mag)^3. The
%%%%%%%%code is faster if 'mag' variable is lower, but the crack fidelity is also reduced
mag = 1/8;%reduce by a factor of mag

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%USER INPUTS IN FIGURE%%%%%%%%%%%%%%%%
fig = figure();
PATCH_3Darray(bondfull,'sym');
title('Orignal 3D image')
d = datacursormode(fig);
datacursormode on 

loop = 1;
while loop < 4
    s = sprintf('click on pixel %d and press enter on keyboard when the coordinate data appears', loop);
    title(s)
    pause
    point1 = getCursorInfo(d);
    if ~isempty(point1)
        pos(loop,:) = point1.Position(1:3);   
%         sprintf('Pixel %d position data recorded at [%f,%f,%f]',loop,pos(loop,1),pos(loop,2),pos(loop,3))     
        loop = loop + 1;
        figure(fig)
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%graph cut algorithm
n = round(n*mag);
bond = VoxelSizeReduction(bondfull,n);
Nnodes = n^3;
Nclasses = 2;%Nclasses = 2 implies one center crack
class = ones(1,Nnodes);%initial classes

%define crack plane through three points 
p1 =pos(1,:).*mag;
p2 =pos(2,:).*mag;
p3 =pos(3,:).*mag;

z1 = p2-p1;
z2 = p3-p1;
su = cross(z1,z2);
D = dot(su,p1);
A=su(1);
B=su(2);
C=su(3);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%develop unary as a scaled distance from the line
fac = sqrt(A^2 + B^2 + C^2);
mat = zeros([n n n]);
mat2 = mat;
dist1 = zeros(Nnodes,1);
for i = 1:Nnodes
   a  = nodetoindex(i,n);
   dist1(i) = (A*a(1) + B*a(2) + C*a(3) - D)/fac;
   mat2(a(1),a(2),a(3)) = (sign(dist1(i))+1)/2;
end
[~,I] = find(dist1 >= 0);
zplus = max(abs(dist1(I)));
[~,I] = find(dist1 < 0);
zminus = max(abs(dist1(I)));

%%%%%%unary cost def%%%%%%%%%%
unary = zeros(2,Nnodes);
unplus = zeros(1,Nnodes);
unminus = zeros(1,Nnodes);
lambda = maxlevels;%controls data term
parfor i = 1:Nnodes
    if dist1(i) >= 0
        unplus(i) = round(lambda*abs(dist1(i))/zplus);
    else
        unminus(i) = round(lambda*abs(dist1(i))/zminus);
    end
end
unary = [unplus;unminus];
%%%%%%%%%%%%%%%%%%%%%%%%%%%


%pairwise cost: just indicate links by bond strength
pairwise = sparse(Nnodes,Nnodes);
parfor i = 1:Nnodes
    a  = nodetoindex(i,n);
    for j = 1:Nnodes
        b  = nodetoindex(j,n);
        if norm(a-b) == 1
            pairwise(i,j) = ((bond(a(1),a(2),a(3)) + bond(b(1),b(2),b(3))));
        end
    end
end

%smooth cost - 0-0 and 1-1 are weighted zero, 0-1 neighbor is weighted 1
parfor i = 1:Nclasses
    for j =1:Nclasses
        labelcost(i,j) = ((i-j)^2);
    end
end

warning off
h = GCO_Create(Nnodes,Nclasses);             % Create new object with NumSites=4, NumLabels=3
GCO_SetDataCost(h,unary);  
GCO_SetSmoothCost(h,labelcost);    
GCO_SetNeighbors(h,pairwise);
GCO_Expansion(h);                % Compute optimal labeling via swap 
disp = GCO_GetLabeling(h);
[E D S] = GCO_ComputeEnergy(h);   % Energy = Data Energy + Smooth Energy
GCO_Delete(h);                   % Delete the GCoptimization object when finished

for l = 1:Nnodes
    a = nodetoindex(l,n);
    mat(a(1),a(2),a(3)) = disp(l);
end

%%%%%%%%%%%%plot on unmagnified image
nfull = length(bondfull);
mat = round(VoxelSizeReduction(mat,nfull));

figure()
PATCH_3Darray(double(bondfull).*(mat-2)./(mat-2),'sym');
xlabel('x');ylabel('y');zlabel('z');xlim([1 nfull]);ylim([1 nfull]);zlim([1 nfull]);
title('Portion 1')


figure()
PATCH_3Darray(double(bondfull).*(mat-1)./(mat-1),'sym');
xlabel('x');ylabel('y');zlabel('z');xlim([1 nfull]);ylim([1 nfull]);zlim([1 nfull]);
title('Portion 2')

end

function [new_vox] = VoxelSizeReduction(old_vox,new_size)
% Function to return voxel position based on node
%
n=size(old_vox,1);
vox_temp = zeros([new_size n n]);
for i = 1 : n
   temp_image(:,:) = old_vox(:,:,i);
   vox_temp(:,:,i) = imresize(temp_image,[new_size n]);
end
new_vox = zeros([new_size new_size new_size]);
for i = 1 : new_size
   temp_image(:,:) = vox_temp(i,:,:);
   new_vox(i,:,:) = imresize(temp_image,[new_size new_size]);
end

end


function a  = nodetoindex(n,imgsz)

n=n-1;
a=[0 0 0];
val = rem(n,imgsz^2);
a(3) = (n-val)/(imgsz^2);
n = n - a(3) *(imgsz^2);
val = rem(val,imgsz);
a(2) = (n-val)/imgsz;
n=n-a(2)*imgsz;
a(1) = n;
a=a+1;
end
